@@ -140,7 +140,7 @@ interp_jmp_buf;
 /* If this is true, for each byte code that is executed, we print on
    stdout the byte index within the current gst_compiled_method and a
    decoded interpretation of the byte code.  */
-mst_Boolean _gst_execution_tracing = false;
+int _gst_execution_tracing = 0;
 
 /* When this is true, and an interrupt occurs (such as SIGABRT),
    Smalltalk will terminate itself by making a core dump (normally it
@@ -152,7 +152,8 @@ mst_Boolean _gst_make_core_file = false;
 mst_Boolean _gst_non_interactive = true;
 
 /* The table of functions that implement the primitives.  */
-static prim_table_entry _gst_primitive_table[NUM_PRIMITIVES];
+prim_table_entry _gst_primitive_table[NUM_PRIMITIVES];
+prim_table_entry _gst_default_primitive_table[NUM_PRIMITIVES];
 
 /* Some performance counters from the interpreter: these
    count the number of special returns.  */
@@ -276,7 +277,7 @@ static interp_jmp_buf *reentrancy_jmp_buf = NULL;
 
 /* when this flag is on and execution tracing is in effect, the top of
    the stack is printed as well as the byte code */
-static mst_Boolean verbose_exec_tracing = false;
+static int verbose_exec_tracing = false;
 
 /* This is the bridge to the primitive operations in the GNU Smalltalk
    system.  This function invokes the proper primitive_func with the
@@ -350,10 +351,13 @@ static void sleep_process (OOP processOOP);
    processes for PROCESSOOP's priority.  Return PROCESSOOP.  */
 static OOP activate_process (OOP processOOP);
 
+/* Restore the virtual machine's state from the ContextPart OOP.  */
+static void resume_suspended_context (OOP oop);
+
 /* Save the virtual machine's state into the suspended Process and
    ContextPart objects, and load them from NEWPROCESS and from
    NEWPROCESS's suspendedContext.  The Processor (the only instance
-   of ProcessorScheduler is also updated accordingly.  */
+   of ProcessorScheduler) is also updated accordingly.  */
 static void change_process_context (OOP newProcess);
 
 /* Mark the semaphores attached to the process system (asynchronous
@@ -470,11 +474,6 @@ static inline void prepare_context (gst_context_part context,
    status (ip, sp, _gst_this_method, _gst_self, ...).  */
 static void unwind_context (void);
 
-/* Used to help minimize the number of primitives used to control the
-   various debugging flags, this routine maps the variable's INDEX to the
-   address of a boolean debug flag, which it returns.  */
-static inline mst_Boolean *bool_addr_index (int index) ATTRIBUTE_PURE;
-
 /* Check whether it is true that sending SENDSELECTOR to RECEIVER
    accepts NUMARGS arguments.  Note that the RECEIVER is only used to
    do a quick check in the method cache before examining the selector
@@ -1320,7 +1319,6 @@ _gst_make_block_closure (OOP blockOOP)
 void
 change_process_context (OOP newProcess)
 {
-  gst_method_context thisContext;
   OOP processOOP;
   gst_process process;
   gst_processor_scheduler processor;
@@ -1351,7 +1349,7 @@ change_process_context (OOP newProcess)
       disable_preemption = !IS_NIL (process->interrupts) &&
 		           TO_INT (process->interrupts) < 0;
 
-      _gst_this_context_oop = process->suspendedContext;
+      resume_suspended_context (process->suspendedContext);
 
       /* Interrupt-enabling cannot be controlled globally from Smalltalk,
          but only on a per-Process basis.  You might think that this leaves
@@ -1376,8 +1374,15 @@ change_process_context (OOP newProcess)
             _gst_enable_interrupts();
 	}
     }
+}
 
-  thisContext = (gst_method_context) OOP_TO_OBJ (_gst_this_context_oop);
+void
+resume_suspended_context (OOP oop)
+{
+  gst_method_context thisContext;
+
+  _gst_this_context_oop = oop;
+  thisContext = (gst_method_context) OOP_TO_OBJ (oop);
   SET_THIS_METHOD (thisContext->method, GET_CONTEXT_IP (thisContext));
   sp = thisContext->contextStack + TO_INT (thisContext->spOffset);
 
@@ -1387,6 +1392,7 @@ change_process_context (OOP newProcess)
 
   _gst_temporaries = thisContext->contextStack;
   _gst_self = thisContext->receiver;
+  free_lifo_context = lifo_contexts;
 }
 
 
@@ -1456,8 +1462,8 @@ add_first_link (OOP semaphoreOOP,
     sem->lastLink = processOOP;
 }
 
-void
-suspend_process (OOP processOOP)
+static void
+remove_process_from_list (OOP processOOP)
 {
   gst_semaphore sem;
   gst_process process, lastProcess;
@@ -1498,6 +1504,12 @@ suspend_process (OOP processOOP)
     }
 
   process->nextLink = _gst_nil_oop;
+}
+
+void
+suspend_process (OOP processOOP)
+{
+  remove_process_from_list (processOOP);
   if (get_scheduled_process() == processOOP)
     ACTIVE_PROCESS_YIELD ();
 }
@@ -1679,11 +1691,25 @@ resume_process (OOP processOOP,
 
   /* 2002-19-12: tried get_active_process instead of get_scheduled_process.  */
   activeOOP = get_active_process ();
-  if (processOOP == activeOOP)
-    return (true);
-
   active = (gst_process) OOP_TO_OBJ (activeOOP);
   process = (gst_process) OOP_TO_OBJ (processOOP);
+  priority = TO_INT (process->priority);
+
+  /* As a special exception, don't preempt a process that has disabled
+     interrupts. ### this behavior is currently disabled.  */
+  ints_enabled = IS_NIL (active->interrupts)
+	         || TO_INT(active->interrupts) <= 0;
+
+  /* resume_process is also used when changing the priority of a ready/active
+     process.  In this case, first remove the process from its current list.  */
+  if (processOOP == activeOOP)
+    {
+      assert (!alwaysPreempt);
+      remove_process_from_list (processOOP);
+    }
+  else if (priority >= TO_INT (active->priority) /* && ints_enabled */ )
+    alwaysPreempt = true;
+
   if (IS_NIL (processOOP) || is_process_terminating (processOOP))
     /* The process was terminated - nothing to resume, fail */
     return (false);
@@ -1695,17 +1721,10 @@ resume_process (OOP processOOP,
       return (true);
     }
 
-  /* As a special exception, don't preempt a process that has disabled
-     interrupts. ### this behavior is currently disabled.  */
-  ints_enabled = IS_NIL (active->interrupts)
-	         || TO_INT(active->interrupts) <= 0;
-
-  priority = TO_INT (process->priority);
   processLists = GET_PROCESS_LISTS ();
   processList = ARRAY_AT (processLists, priority);
 
-  if ((priority >= TO_INT (active->priority) /* && ints_enabled */ )
-      || alwaysPreempt)
+  if (alwaysPreempt)
     {
       /* We're resuming a process with a *equal or higher* priority, so sleep
          the current one and activate the new one */
@@ -1713,11 +1732,13 @@ resume_process (OOP processOOP,
       activate_process (processOOP);
     }
   else
-    /* this process has a lower priority than the active one, so the
-       policy is that it doesn't preempt the currently running one.
-       Anyway, it must be the first in its priority queue - so don't
-       put it to sleep.  */
-    add_first_link (processList, processOOP);
+    {
+      /* this process has a lower priority than the active one, so the
+         policy is that it doesn't preempt the currently running one.
+         Anyway, it must be the first in its priority queue - so don't
+         put it to sleep.  */
+      add_first_link (processList, processOOP);
+    }
 
   return (true);
 }
@@ -1960,8 +1981,6 @@ semaphore_new (int signals)
   return (semaphoreOOP);
 }
 
-/* runs before every evaluation (_gst_execute_statements) and before GC turned on.
-   Note that we don't use the incubator because _gst_processor_oop is a global.  */
 void
 _gst_init_process_system (void)
 {
@@ -2026,25 +2045,66 @@ create_callin_process (OOP contextOOP)
   return (initialProcessOOP);
 }
 
-mst_Boolean *
-bool_addr_index (int index)
+int
+_gst_get_var (enum gst_var_index index)
 {
   switch (index)
     {
-    case 0:
-      return (&_gst_declare_tracing);
-    case 1:
-      return (&_gst_execution_tracing);
-    case 2:
-      return (&verbose_exec_tracing);
-    case 3:
-      return (&_gst_gc_message);
+    case GST_DECLARE_TRACING:
+      return (_gst_declare_tracing);
+    case GST_EXECUTION_TRACING:
+      return (_gst_execution_tracing);
+    case GST_EXECUTION_TRACING_VERBOSE:
+      return (verbose_exec_tracing);
+    case GST_GC_MESSAGE:
+      return (_gst_gc_message);
+    case GST_VERBOSITY:
+      return (_gst_verbosity);
+    case GST_MAKE_CORE_FILE:
+      return (_gst_make_core_file);
+    case GST_REGRESSION_TESTING:
+      return (_gst_regression_testing);
     default:
-      return (NULL);		/* index out of range, signal the error 
-				 */
+      return (-1);
     }
 }
 
+int
+_gst_set_var (enum gst_var_index index, int value)
+{
+  int old = _gst_get_var (index);
+  if (value < 0)
+    return -1;
+
+  switch (index)
+    {
+    case GST_DECLARE_TRACING:
+      _gst_declare_tracing = value;
+      break;
+    case GST_EXECUTION_TRACING:
+      _gst_execution_tracing = value;
+      break;
+    case GST_EXECUTION_TRACING_VERBOSE:
+      verbose_exec_tracing = value;
+      break;
+    case GST_GC_MESSAGE:
+      _gst_gc_message = value;
+      break;
+    case GST_VERBOSITY:
+      _gst_verbosity = value;
+      break;
+    case GST_MAKE_CORE_FILE:
+      _gst_make_core_file = value;
+      break;
+    case GST_REGRESSION_TESTING:
+      _gst_regression_testing = true;
+      break;
+    default:
+      return (-1);
+    }
+
+  return old;
+}
 
 
 void
@@ -2133,25 +2193,6 @@ _gst_nvmsg_send (OOP receiver,
   currentProcessOOP = get_active_process ();
   change_process_context (processOOP);
 
-#if 0
-  if (_gst_kernel_initialized)
-    {
-      gst_array argsArray;
-      OOP argsArrayOOP;
-      argsArray = new_instance_with (_gst_array_class, numArgs, &argsArrayOOP);
-      for (i = 0; i < sendArgs; i++)
-        argsArray->data[i] = args[i];
-
-      INC_ADD_OOP (argsArrayOOP);
-      dirMessageOOP = _gst_directed_message_new_args (receiver, sendSelector, argsArrayOOP);
-
-      receiver = _gst_process_class;
-      sendSelector = _gst_start_execution_symbol;
-      sendArgs = 1;
-      args = &dirMessageOOP;
-    }
-#endif
-
   PUSH_OOP (receiver);
   for (i = 0; i < sendArgs; i++)
     PUSH_OOP (args[i]);
@@ -2468,7 +2509,7 @@ interrupt_handler (int sig)
 #ifdef HAVE_EXECINFO_H
       /* Don't print a backtrace, for example, if exiting during a
 	 compilation.  */
-      if ((_gst_verbosity > 2 && (ip || _gst_gc_running))
+      if ((_gst_verbosity == 3 && (ip || _gst_gc_running))
 	  || is_serious_error || sig == SIGUSR1)
 	{
           PTR array[11];
@@ -2626,7 +2667,16 @@ execute_primitive_operation (int primitive, volatile int numArgs)
 prim_table_entry *
 _gst_get_primitive_attributes (int primitive)
 {
-  return &_gst_primitive_table[primitive];
+  return &_gst_default_primitive_table[primitive];
+}
+
+void
+_gst_set_primitive_attributes (int primitive, prim_table_entry *pte)
+{
+  if (pte)
+    _gst_primitive_table[primitive] = *pte;
+  else
+    _gst_primitive_table[primitive] = _gst_default_primitive_table[0];
 }
 
 void
